@@ -45,7 +45,7 @@
 #include <openobex/obex_const.h>
 
 #include "plugin.h"
-#include "logging.h"
+#include "log.h"
 #include "obex.h"
 #include "mimetype.h"
 #include "service.h"
@@ -138,6 +138,7 @@ static void *filesystem_open(const char *name, int oflag, mode_t mode,
 	char *folder;
 	gboolean root;
 	int fd = open(name, oflag, mode);
+	uint64_t avail;
 
 	if (fd < 0) {
 		if (err)
@@ -181,7 +182,8 @@ static void *filesystem_open(const char *name, int oflag, mode_t mode,
 	if (size == NULL)
 		goto done;
 
-	if (buf.f_bsize * buf.f_bavail < *size) {
+	avail = (uint64_t) buf.f_bsize * buf.f_bavail;
+	if (avail < *size) {
 		if (err)
 			*err = -ENOSPC;
 		goto failed;
@@ -235,7 +237,7 @@ struct capability_object {
 	int pid;
 	int output;
 	int err;
-	unsigned int watch;
+	gboolean aborted;
 	GString *buffer;
 };
 
@@ -246,6 +248,19 @@ static void script_exited(GPid pid, int status, void *data)
 
 	object->pid = -1;
 
+	DBG("pid: %d status: %d", pid, status);
+
+	g_spawn_close_pid(pid);
+
+	/* free the object if aborted */
+	if (object->aborted) {
+		if (object->buffer != NULL)
+			g_string_free(object->buffer, TRUE);
+
+		g_free(object);
+		return;
+	}
+
 	if (WEXITSTATUS(status) != EXIT_SUCCESS) {
 		memset(buf, 0, sizeof(buf));
 		if (read(object->err, buf, sizeof(buf)) > 0)
@@ -253,8 +268,6 @@ static void script_exited(GPid pid, int status, void *data)
 		obex_object_set_io_flags(data, G_IO_ERR, -EPERM);
 	} else
 		obex_object_set_io_flags(data, G_IO_IN, 0);
-
-	g_spawn_close_pid(pid);
 }
 
 static int capability_exec(const char **argv, int *output, int *err)
@@ -270,6 +283,8 @@ static int capability_exec(const char **argv, int *output, int *err)
 		return -EPERM;
 	}
 
+	DBG("executing %s pid %d", argv[0], pid);
+
 	return pid;
 }
 
@@ -314,10 +329,8 @@ static void *capability_open(const char *name, int oflag, mode_t mode,
 	if (object->pid < 0)
 		goto fail;
 
-	object->watch = g_child_watch_add(object->pid, script_exited, object);
-
-	if (size)
-		*size = 1;
+	/* Watch cannot be removed while the process is still running */
+	g_child_watch_add(object->pid, script_exited, object);
 
 done:
 	if (err)
@@ -397,7 +410,7 @@ static GString *append_listing(GString *object, const char *name,
 			ret = lstat(fullname, &fstat);
 
 		if (ret < 0) {
-			debug("%s: %s(%d)", root ? "stat" : "lstat",
+			DBG("%s: %s(%d)", root ? "stat" : "lstat",
 					strerror(errno), errno);
 			g_free(filename);
 			g_free(fullname);
@@ -510,19 +523,29 @@ static ssize_t capability_read(void *object, void *buf, size_t count,
 static int capability_close(void *object)
 {
 	struct capability_object *obj = object;
+	int err = 0;
 
-	if (obj->pid >= 0) {
-		g_source_remove(obj->watch);
-		kill(obj->pid, SIGTERM);
-		g_spawn_close_pid(obj->pid);
+	if (obj->pid < 0)
+		goto done;
+
+	DBG("kill: pid %d", obj->pid);
+	err = kill(obj->pid, SIGTERM);
+	if (err < 0) {
+		err = -errno;
+		error("kill: %s (%d)", strerror(-err), -err);
+		goto done;
 	}
 
+	obj->aborted = TRUE;
+	return 0;
+
+done:
 	if (obj->buffer != NULL)
 		g_string_free(obj->buffer, TRUE);
 
 	g_free(obj);
 
-	return 0;
+	return err;
 }
 
 static struct obex_mime_type_driver file = {
@@ -535,6 +558,7 @@ static struct obex_mime_type_driver file = {
 
 static struct obex_mime_type_driver capability = {
 	.target = FTP_TARGET,
+	.target_size = TARGET_SIZE,
 	.mimetype = "x-obex/capability",
 	.open = capability_open,
 	.close = capability_close,
@@ -543,6 +567,7 @@ static struct obex_mime_type_driver capability = {
 
 static struct obex_mime_type_driver folder = {
 	.target = FTP_TARGET,
+	.target_size = TARGET_SIZE,
 	.mimetype = "x-obex/folder-listing",
 	.open = folder_open,
 	.close = string_free,
@@ -551,6 +576,7 @@ static struct obex_mime_type_driver folder = {
 
 static struct obex_mime_type_driver pcsuite = {
 	.target = FTP_TARGET,
+	.target_size = TARGET_SIZE,
 	.who = PCSUITE_WHO,
 	.who_size = PCSUITE_WHO_SIZE,
 	.mimetype = "x-obex/folder-listing",
